/*
 *   MediaMVP Server Library
 *
 *   (C) 2003 Dominic Morris
 *
 *   $Id: url_http.c,v 1.3 2004/01/24 18:47:54 dom Exp $
 *   $Date: 2004/01/24 18:47:54 $
 *
 *
 *   Deals with opening a http connection and fobbing off playlist detection
 *   to a lower level
 */

#include "libmvp_internal.h"


static void           *http_open(char *name, int *type, fops_t **fops_ptr, readready_cb cb, void *cb_ptr );
static void            http_close(void *ptr);
static off_t           http_seek(void *ptr, off_t offset, int whence);
static int             http_read(void *ptr, unsigned char *buf, size_t buflen);

static void            http_read_event(int fd, short event, void *arg);
static int             http_open_direct(char *url);

typedef struct {
    struct  event *ev;
    readready_cb   cb;
    void          *cb_ptr;
    char          *location;
    int            fd;
} httpfile_t;

static fops_t          http_fops = { 
    NULL,
    http_open,
    http_close,
    http_seek,
    http_read
};


#define ACCEPT "Accept: video/mpeg, video/x-mpegurl, */*\r\n"

void urlhttp_init()
{
    urlhandler_add("mpeg;http://",&http_fops);
    urlhandler_add("mp3;http://",&http_fops);
    urlhandler_add("http://",&http_fops);
}

static void *http_open(char *url, int *type,fops_t **fops_ptr,readready_cb cb, void *cb_ptr )
{
    httpfile_t   *httpfile;
    void         *ptr;
    int           fd;
    int           ret;
    bool_t        playlist = FALSE;

    /* This is a bit of splendiferous hack <grin> */
    if ( strncmp(url,"mpeg;",5) == 0 ) {
        *type = MEDIA_MPEG | MEDIA_LIVE;
        url += 5;
    } else if ( strncmp(url,"mp3;",3) == 0 ) {
        *type = MEDIA_MP3 | MEDIA_LIVE;
        url += 4;
    } else {       /* Unknown type, try to guess from the file extension */
        char    *ptr;

        playlist = TRUE;
        if ( (ptr = strrchr(url,'.') ) != NULL ) {    
            printf("%s %s\n",url,ptr);
            if ( strcasecmp(ptr,".mp3") == 0 ) {
                *type = MEDIA_MP3;
                playlist = FALSE;
            } else if ( strcasecmp(ptr,".mpeg") == 0 || strcasecmp(ptr,".vdr") == 0 || strcasecmp(ptr,".mpg") == 0 ) {
                *type = MEDIA_MPEG;
                playlist = FALSE;
            }
        }
    }

    if ( (fd = http_open_direct(url) ) == -1 ) {
        return NULL;
    }

    /* Createa a http structure */
    httpfile = calloc(1,sizeof(*httpfile));
    httpfile->fd = fd;
    httpfile->cb_ptr = cb_ptr;
    httpfile->cb = cb;

    if ( playlist == TRUE && (ret = playlist_check("",fd,httpfile,type,&ptr,fops_ptr,cb,cb_ptr) ) != 0 ) {
        /* This is where the playlist has done something, be it a bad filetype or a playlist, we needn't
           care about it, we know that it has closed up our file for us, so return whatever ptr was set 
        */
        return ptr;     
    } 
  
    /* Indicate that it's live */
//    *type |= MEDIA_SOCKET;
  
    /* Now we must set up an event to chuck things into the system */
    if ( *type & MEDIA_LIVE ) {
        httpfile->ev = calloc(1,sizeof(*httpfile->ev));
        event_set(httpfile->ev,fd,EV_READ,http_read_event,httpfile);
        event_add(httpfile->ev,NULL);
    }
    return httpfile;
}

static void http_read_event(int fd, short event, void *arg)
{
    httpfile_t  *file = arg;

    /* Readd the event here */
    event_add(file->ev,NULL);

    file->cb(file->cb_ptr);
}

static void http_close(void *ptr)
{
    httpfile_t   *file = ptr;

    if ( file->ev ) {
        event_del(file->ev);
        file->ev = NULL;
    }
    close(file->fd);
    shutdown(file->fd,SHUT_RDWR);

    free(file);
}

static off_t http_seek(void *ptr,off_t offset, int whence)
{
    return (off_t) -1;
}

static int http_read(void *ptr, unsigned char *buf, size_t bufsize)
{
    httpfile_t  *file = ptr;

    return read(file->fd,buf,bufsize);
}





/* HTTP stuff is stolen from mpegtools/ctools.c */
static void write_all (int fd, uint8_t *data, int length)
{
        int r;

        while (length) {
                if ((r = write(fd, data, length)) > 0) {
                        data += r;
                        length -= r;
                }
        }
}

static int read_all (int fd, uint8_t *data, int length)
{
        int c = 0;

        while(c<length) {
                if( read(fd, data+c, 1) == 1) {
                        c++;
                        if(data[c-1] == '\n') {
                                data[c] = 0;
                                break;
                        }
                }
                else {
                        return -1;
                }
        }
        return 0;
}

static char *url2host (uint8_t *url, char **name, uint32_t *ip, uint32_t *port)
{
        uint8_t *murl;
        struct hostent *hoste;
        struct in_addr haddr;
        int found_ip = 1;
        
        if (!(strncmp(url, "http://", 7)))
                url += 7;

        *name = strdup(url);
        if (!(*name)) {
                *name = NULL;
                return (NULL);
        }

        murl = url;
        while (*murl && *murl != ':' && *murl != '/') {
                if ((*murl < '0' || *murl > '9') && *murl != '.')
                        found_ip = 0;
                murl++;
        }
        (*name)[murl - url] = 0;
        if (found_ip) {
                if ((*ip = inet_addr(*name)) == INADDR_NONE)
                        return (NULL);
        } else {
                if (!(hoste = gethostbyname(*name)))
                        return (NULL);
                memcpy (&haddr, hoste->h_addr, sizeof(haddr));
                *ip = haddr.s_addr;
        }

        if (!*murl || *murl == '/') {
                *port = 80;
                return (murl);
        }
        *port = atoi(++murl);

        while (*murl && *murl != '/')
                murl++;
        return (murl);
}



static int http_open_direct(char *url)
{
    char purl[1024], *host, req[1024], *sptr;
    uint32_t ip;
    uint32_t port;
    int sock;
    int reloc, relocnum = 0;
    struct sockaddr_in server;
    int mfd;


  
    strncpy (purl, url, 1023);
    purl[1023] = '\0';

    do {
        host = NULL;
        strcpy (req, "GET ");
        if (!(sptr = url2host(purl, &host, &ip, &port))) {
            Dprintf(ERROR,"Unknown host %s\n",purl);
            return -1;
        }
        strcat (req, sptr);
        sprintf (req + strlen(req),
                 " HTTP/1.0\r\nUser-Agent: %s/%s\r\n",
                 SERVER_NAME, SERVER_VERSION);
        if (host) {
            sprintf(req + strlen(req),
                    "Host: %s:%u\r\n", host, port);
            free (host);
        }

        strcat (req, ACCEPT);
        strcat (req, "\r\n");
                
        server.sin_port = htons(port);
        server.sin_family = AF_INET;
        server.sin_addr.s_addr = ip;

        if ((sock = socket(PF_INET, SOCK_STREAM, 6)) < 0) {
            return -1;  
        }

        if (connect(sock, (struct sockaddr *)&server, 
                    sizeof(server))) {
            close(mfd);
            Dprintf(ERROR,"Couldn't connect to %s\n",purl);
            return -1;                
        }
                
        write_all (sock, req, strlen(req));
        if (!(mfd = fileno(fdopen(sock, "rb")))) {
            close(mfd);
            shutdown(mfd,SHUT_RDWR);
            return -1;    
        }
        reloc = 0;
        purl[0] = '\0';
        if ( read_all (mfd, req, 1023) == -1 ) {
            return -1;
        }
        if ((sptr = strchr(req, ' '))) {
            switch (sptr[1]) {
            case '2':
                break;
            case '3':
                reloc = 1;
            default:
                Dprintf(ERROR, "HTTP req failed:%s",sptr+1);
                close(mfd);
                shutdown(mfd,SHUT_RDWR);
                return -1;             
            }
        }
        do {
            if ( read_all (mfd,req, 1023) == -1 ) {
                return -1;
            }
            if (!strncmp(req, "Location:", 9))
                strncpy (purl, req+10, 1023);
        } while (req[0] != '\r' && req[0] != '\n');
    } while (reloc && purl[0] && relocnum++ < 3);
    if (reloc) {
        Dprintf(ERROR,"Too many HTTP relocations.\n");
        close(mfd);
        shutdown(mfd,SHUT_RDWR);
        return -1;
    }
  
    return sock;
}


